#include <unistd.h>
#include <stdio.h>
#include <malloc.h>
#include <alloca.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/socket.h>
#include <resolv.h>
#include <netdb.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include "crt.h"
#include "sockcommon.h"
#include "dane-support.h"

void *usage(bool iserr) {
  FILE *out = iserr ? stderr : stdout;
  fprintf(out,"atea tii [-cqsdn0] [options] [--] url1 [options] [--] url2 ... download (get) the specified urls\n"); 
  fprintf(out,"     --cert xy.pem ... manually specify the server certificate\n");
  fprintf(out,"     --dane ... dynamically get the server certificate via dane\n");
  fprintf(out,"     --no-check-cert / -n ... connect via https without checking the certificate\n");
  fprintf(out,"     -0 ... default check: dane for https and no check for http\n");
  fprintf(out,"     -6 ... force usage of ipv6\n");
  fprintf(out,"     -c / -q ... continue an interrupted download / be quiet\n");
  fprintf(out,"     --to-stdout / -s / --to xy.dat ... print the output to stdout / into xy.dat (overwrites!)\n");
  fprintf(out,"atea tuu file1 file2 url1,2 ... upload file1 to url1/file1 (not implemented)\n");
  fprintf(out,"  supported protocols: http, https\n\n");
  exit(iserr?1:0);
}

bool InitCTX(SSL_CTX **ctx) {
    const SSL_METHOD *method;
    OpenSSL_add_all_algorithms();  /* Load cryptos, et.al. */
    SSL_load_error_strings();   /* Bring in and register error messages */
    method = TLS_client_method();  /* Create new client-method instance */
    //method = TLSv1_2_client_method();  /* Create new client-method instance, old deprecated call */
    *ctx = SSL_CTX_new(method);   /* Create new context */
    if ( *ctx == NULL ) { ERR_print_errors_fp(stderr); return false; }
    return true;
}

bool wasa_sigwinch = true, progress_drawed = false; uint64 dld_start_pos, dld_last_drawn_pos; clockid_t clockid = 0;
#define ProgressTraceSize 5
int progressTraceLen, ptIdx;
struct DldState { struct timespec time; uint64 dld_pos; } progressTrace[ProgressTraceSize];
struct DldState lastProgressState;

void InitProgress( uint64 dld_pos, uint64 content_length) {
  int ret;
  dld_start_pos = dld_pos; lastProgressState.dld_pos = dld_pos; progress_drawed = false;
  if(!clockid) {
    clockid = CLOCK_MONOTONIC; ret = clock_gettime(clockid,&lastProgressState.time);
    if(ret) { clockid = CLOCK_REALTIME; clock_gettime(clockid,&lastProgressState.time); }
  } else clock_gettime(clockid,&lastProgressState.time); 
  ptIdx=0; progressTraceLen=1; progressTrace[0] = lastProgressState;
}

void getMeasure( int *value, char *measure, uint64 rawval ) {
  if( rawval < 1000 ) { *value = rawval; *measure = ' '; return; }
  rawval = ( rawval + 500 ) / 1000;
  if( rawval < 1000 ) { *value = rawval; *measure = 'K'; return; }
  rawval = ( rawval + 500 ) / 1000;
  if( rawval < 1000 ) { *value = rawval; *measure = 'M'; return; }
  rawval = ( rawval + 500 ) / 1000;
  if( rawval < 1000 ) { *value = rawval; *measure = 'G'; return; }
  rawval = ( rawval + 500 ) / 1000;
  *value = rawval; *measure = 'T';
}

void UpdateProgress( uint64 dld_pos, uint64 content_length) {
  int li, ci; dbl_int timediff; int slidersz, predldd, prevdldd, dldd; int prev_termCols; int i; 
  uint64 speed, med_speed; int speed_value; char speed_measure; uint64 mins, hours, days;
  li = ptIdx; ptIdx++; if(ptIdx>=ProgressTraceSize) ptIdx=0; progressTraceLen = min_macro(progressTraceLen+1,ProgressTraceSize);
  clock_gettime(clockid,&progressTrace[ptIdx].time); progressTrace[ptIdx].dld_pos = dld_pos;
  timediff = get_timediff_ms(&progressTrace[li].time,&progressTrace[ptIdx].time);
  fprintf(stderr,"%llu\n",timediff);
  if( timediff < 500 ) return;
  if(wasa_sigwinch) { wasa_sigwinch=false; prev_termCols=termCols; queryWinSz(); if(termCols!=prev_termCols) progress_drawed=false; }
  // len("100% 365d,12:23min 500Gb/s ") = 27
  slidersz = termCols - 27 - 4;
  if( !progress_drawed && slidersz > 0 ) {
    fprintf(stderr,"\r |");
    predldd = ( dld_start_pos * slidersz ) / content_length;
    dldd = ( dld_pos * slidersz ) / content_length;
    for(i=0;i<predldd;i++) fputc('+',stderr);
    for(;i<dldd;i++) fputc('=',stderr); 
    for(;i<slidersz;i++) fputc('-',stderr); 
    fprintf(stderr,"| ");
    progress_drawed=true; dld_last_drawn_pos = dld_pos;
  } else {
    prevdldd = ( dld_last_drawn_pos * slidersz ) / content_length;
    dldd = ( dld_pos * slidersz ) / content_length;
    if( prevdldd < dldd ) {
      fprintf(stderr,"\r" e_moveRight "=", prevdldd+2 );
      for(i=prevdldd+1; i<dldd; i++) fputc('=',stderr);
      dld_last_drawn_pos = dld_pos;
  } }
  fprintf(stderr,"\r" e_moveRight "%3i%% ", slidersz+4, (int)( dld_pos * 100 / content_length ) );
  speed = ( dld_pos - progressTrace[li].dld_pos ) * 8000 / timediff;

  ci = ( ptIdx + ProgressTraceSize - progressTraceLen + 1 ) % ProgressTraceSize;
  timediff = get_timediff_ms(&progressTrace[ci].time,&progressTrace[ptIdx].time);
  med_speed = ( dld_pos - progressTrace[ci].dld_pos ) * 1000 / timediff;
  mins = ( content_length - dld_pos ) / ( med_speed * 60 );
  hours = mins / 60; mins -= hours * 60;
  days = hours / 24; hours -= days * 24;
  if(days>999) fprintf(stderr,">999days ");
  else if(days>0) fprintf(stderr,"%id,%2i:%2imin ", (int)days, (int)hours, (int)mins );
  else fprintf(stderr,"%2i:%2imin ", (int)hours, (int)mins );

  getMeasure( &speed_value, &speed_measure, speed );
  if(speed_value>999) fprintf(stderr,">mx%cb/s" e_clearEOL, speed_measure );
  else fprintf(stderr,"%3i%cb/s" e_clearEOL, speed_value, speed_measure ); 
  fflush(stderr);
}

void ShowCerts(SSL* ssl) {
    X509 *cert; char *line;
    cert = SSL_get_peer_certificate(ssl); /* get the server's certificate */
    if ( cert != NULL ) {
        printf("Server certificates:\n");
        line = X509_NAME_oneline(X509_get_subject_name(cert), 0, 0);
        printf("Subject: %s\n", line);
        free(line);       /* free the malloc'ed string */
        line = X509_NAME_oneline(X509_get_issuer_name(cert), 0, 0);
        printf("Issuer: %s\n", line);
        free(line);       /* free the malloc'ed string */
        X509_free(cert);     /* free the malloc'ed certificate copy */
    }
    else
        printf("Info: No client certificates configured.\n");
}

X509 *required_cert = NULL;  // given by --cert parameter

#define DIGESTBUF_LEN EVP_MAX_MD_SIZE
X509* read_cert_from_file( const char *filename ) {
  int result; X509 *cert = NULL; BIO *bo = NULL;
  bo = BIO_new(BIO_s_file());
  result = BIO_read_filename(bo, filename );
  if(!result) { BIO_free(bo); fprintf(stderr,"error reading certfile '%s' - will refuse to download over https.\n",filename); return NULL; }
  PEM_read_bio_X509(bo, &cert, 0, NULL);
  BIO_free(bo);
  return cert;
}

bool X509_compare_equal( X509 *certA, X509 *certB ) {
  int lenA, lenB; unsigned char *bufA = NULL, *bufB = NULL; bool result;
  lenA = i2d_X509(certA,&bufA); lenB = i2d_X509(certB,&bufB);
  result = ( lenA == lenB ) && ( lenA > 0 ) && !memcmp(bufA,bufB,lenA);
  OPENSSL_free(bufA); OPENSSL_free(bufB);
  return result;
}

int fmtCertName( char *buf, int bufsz, const char *hostname, const char *postfix, int i ) {
  if(i<=1) return snprintf(buf,bufsz,"%s%s.pem",hostname,postfix);
  return snprintf(buf,bufsz,"%s%s-%i.pem",hostname,postfix,i);
}

void saveCert( const char *hostname, char *postfix, X509 *cert ) {
  int i = 1, res, bufsz = strlen(hostname)+strlen(postfix)+4+4+1;   // len(".pem") = len("NNNN") = 4
  char filename[bufsz]; BIO *bo; X509 *fileCert = NULL; bool iseq; struct stat statbuf;
  bo = BIO_new(BIO_s_file());
  while( i < 10000 ) {
    fmtCertName(filename,bufsz,hostname,postfix,i);
    res = BIO_read_filename(bo,filename);
    if(res) { 
      PEM_read_bio_X509(bo, &fileCert, 0, NULL);
      iseq = X509_compare_equal(cert,fileCert);
      X509_free(fileCert); fileCert = NULL;
      if(iseq) {
	fprintf(stderr,"server certificate equals the certificate found in '%s'.\n",filename);
	BIO_free(bo); 
	return;
      }
    } else if( lstat(filename,&statbuf) == -1 ) break;
    i++;
  }
  if(i>=10000) fprintf(stderr,"number space for rogue cert file names exhausted.\n");
  else {
    if(!BIO_write_filename(bo,filename)) fprintf(stderr,"error opening '%s' for writing.\n",filename); else {
      PEM_write_bio_X509(bo,cert);
      fprintf(stderr,"server cert written to '%s'.\n",filename);
  } }
  BIO_free(bo); 
}

typedef int(*PWriteOut)(void*,const void *,int);
typedef int(*PReadIn)(void*,void*,int);

int directWrite( void *fd, const void *buf, int len ) { return write(*(int*)fd,buf,len); }
int directRead( void *fd, void *buf, int len ) { return read(*(int*)fd,buf,len); }

#define BUFSZ 4096
char buf[BUFSZ];

int printheaders = 0; 
enum FilenameSpec { FS_IMPLICIT = 0, FS_EXPLICIT = 1, FS_STDOUT = 2 } fileNameMode = FS_IMPLICIT;
const char *fileName = NULL; char *dldFileName; FILE *output = NULL; bool overwriteTargetFile = false, saveErrorResponse = false;

bool OpenOutput( struct UrlSpec *urlspec ) {
  int len, maxlen, fd, num = 1;
  if( fileNameMode == FS_STDOUT ) { output = stdout; dldFileName = NULL; return true; }
  if( fileNameMode == FS_IMPLICIT ) fileName = urlspec->target;
  if( fileName == NULL ) { fprintf(stderr,"error: need to specify target filename with --to for %s.\n",urlspec->fullURL); return false; }
  len = strlen(fileName);
  dldFileName = (char*)malloc(maxlen=len+5+13);   // len(".downloading") == 12
  snprintf(dldFileName,maxlen,"%s.downloading",fileName);
  fd = open(dldFileName,O_CREAT|O_EXCL|O_WRONLY,0666);
  if( fd == -1 ) {
    do {
      snprintf(dldFileName,maxlen,"%s-%i.downloading",fileName,num);
      fd = open(dldFileName,O_CREAT|O_EXCL|O_WRONLY,0666);
      if( fd >= 0 ) break;
      num++;
    } while( num < 10000 );
    if( fd == -1 ) { fprintf(stderr,"ran out of filename space: %s - can not download!\n",dldFileName); return false; }
  }
  output = fdopen(fd,"w"); if(!output) { perror(dldFileName); return false; }
  return true;
}

uint64 dld_pos;

bool AppendOutput( struct UrlSpec *urlspec ) {
  DIR *dirs; struct dirent *dirent; bool result = false;
  int len, entlen; struct stat statbuf;
  if( fileNameMode == FS_STDOUT ) { fprintf(stdout,"can not continue download on stdout; downloading from scratch (%s).\n",urlspec->fullURL); return false; }
  if( fileNameMode == FS_IMPLICIT ) fileName = urlspec->target;
  if( fileName == NULL ) return false;
  dirs = opendir("."); len = strlen(fileName); dldFileName = NULL;
  if(!dirs) perror("trying to list directory content:");
  while(true) {
    dirent = readdir(dirs);
    if(!dirent) break;
    entlen = strlen(dirent->d_name);  // len(".downloading") == 12
    if(!strncmp(dirent->d_name,fileName,len) && entlen >= len+12 && !strncmp(dirent->d_name+entlen-12,".downloading",12) && ( entlen == len+12 || dirent->d_name[len] == '-' ) ) {
      if( stat(dirent->d_name,&statbuf) != -1 ) {
	if( statbuf.st_size > dld_pos && (( statbuf.st_mode & S_IFMT ) == S_IFREG ) ) {   // size + is a regular file
	  free(dldFileName); dldFileName = (char*)malloc(entlen+1);
	  memcpy(dldFileName,dirent->d_name,entlen+1);
	  dld_pos = statbuf.st_size;
	  result = true;
    } } }
  }
  closedir(dirs);
  if( result ) {
    if(verbosity>=1) fprintf(stderr,"continuing download of '%s'.\n",dldFileName);
    output = fopen(dldFileName,"a");
    if(!output) { 
      fprintf(stderr,"error opening '%s' for append: ",dldFileName); perror("");
      free(dldFileName); result = false;
  } }
  return result;
}

void RenameOutput() {
  struct stat statbuf; int num = 1; char *targetName; int maxlen;
  //printf("%s->%s\n",dldFileName,fileName);
  if(!dldFileName) return;
  fclose(output);
  if( overwriteTargetFile || lstat(fileName,&statbuf) == -1 ) {
    if( rename(dldFileName,fileName) != 0 ) { fprintf(stderr,"error renaming %s into ",dldFileName); perror(fileName); }
  } else { // file already exists, find another name
    maxlen = strlen(fileName)+6;
    targetName = (char*)alloca(maxlen);
    do {
      snprintf(targetName,maxlen,"%s.%i",fileName,num);
      if( lstat(targetName,&statbuf) == -1 ) break;
      num++;
    } while( num < 10000 );
    if( num >= 10000 ) fprintf(stderr,"all files exist: %s; not renaming %s.\n",targetName,dldFileName);
    else if( rename(dldFileName,targetName) != 0 ) { fprintf(stderr,"error renaming %s into ",dldFileName); perror(targetName); }
    else fprintf(stderr,"saved as '%s'.\n",targetName);
  }
  free(dldFileName);
}

#define InternalErr 11
#define FileErr 12
#define ConnErr 13

bool do_continue = false, quiet = true;

bool mk_HTTP_Request( struct UrlSpec *urlspec, void *fd, PReadIn readin, PWriteOut writeout, bool isssl ) {
  int len, ioc, i, j, k; char *bufptr; bool last_c_nl; char *stopchar; uint64 content_length = 0; bool had_content_length = false, header_continued, break2;
  bool headerline; long statuscode; int l; bool saveOutput; int errcode = 0; bool is_continued = false, saveOutput_set = false;
  dld_pos = 0;
  len = snprintf(buf,BUFSZ,"GET %s HTTP/1.1\r\n", urlspec->path);
  if(do_continue) {
    is_continued = AppendOutput(urlspec);
    if(is_continued) len += snprintf(buf+len,BUFSZ-len,"Range: bytes=%llu-\r\n",dld_pos); 
  }
  if( urlspec->port_override >= 0 ) len += snprintf(buf+len,BUFSZ-len,"Host: %s:%i",urlspec->host,urlspec->port_override);
  else len += snprintf(buf+len,BUFSZ-len,"Host: %s",urlspec->host);
  len += snprintf(buf+len,BUFSZ-len,"\r\nUser-Agent: atea/0.3" OSSTR "\r\nAccept: */*\r\nConnection: Keep-Alive\r\nAccept-Encoding: identity\r\n\r\n");
  if( len < 1 || len > BUFSZ-2 ) { fprintf(stderr,"internal error at formatting the request\n"); return InternalErr; }
  //fwrite(buf,len,1,stdout);
  bufptr=buf; while( len > 0 ) {
    ioc = writeout(fd,bufptr,len);
    if( ioc <= 0 ) {
      fprintf(stderr,"error writing to socket");
      return ConnErr;
    }
    bufptr+=ioc; len-=ioc;
  }
  bufptr=buf; len=0; header_continued=false; break2=false;
  headerline=true; statuscode=-1;
  while(true) {
    ioc = readin(fd,bufptr,BUFSZ-len);
    if(ioc<0) { fprintf(stderr,"read error"); return ConnErr; }
    if(!ioc) { j=len; break;}
    bufptr+=ioc; len+=ioc;
    i=0; last_c_nl = !header_continued; // usually last_c_nl=true
    do {
      j=i; while( j<len && buf[j] != '\n' ) { last_c_nl &= buf[j]=='\r'; j++; }
      if(j>=len) break;
      if(last_c_nl) { j++; break2=true; break; } else last_c_nl=true;
      if(headerline) {
        k=i; while( k<j && buf[k]!=' ' && buf[k]!='\t' ) k++;
        k++; while( k<j &&(buf[k]==' '||buf[k]=='\t')) k++;
	errno = 0; statuscode = strtol( buf+k, &stopchar, 0 );
        if( !( errno==0 && ( !*stopchar || *stopchar == '\r' || *stopchar == '\n' || *stopchar == ' ' || *stopchar == '\t' ) ) ) statuscode = -1;
	l=k+1; while( l<j && buf[l]!='\r' && buf[l]!='\n' ) l++;
	if( statuscode!=200 && statuscode!=206 ) { 
	  //printheaders = 300 <= statuscode && statuscode < 400;
	  if( !printheaders && statuscode != 416 ) { fwrite(buf+k,l-k,1,stderr); fputc('\n',stderr); }
	  errcode = statuscode/100; if(errcode<2) errcode=2; else if(errcode>9) errcode=2;
	}
	headerline=false;
      } else if( !strncasecmp(buf+i,"Content-Length:",min_macro(15,j-i)) && ( !header_continued || i>0 ) ) { k=i+15; while(buf[k]==' '||buf[k]=='\t') k++;
        errno = 0; content_length = strtoull( buf+k, &stopchar, 0 );
        if( errno==0 && ( !*stopchar || *stopchar == '\r' || *stopchar == '\n' || *stopchar == ' ' || *stopchar == '\t' ) ) had_content_length = true;
      } else if( !strncasecmp(buf+i,"Location:",min_macro(9,j-i)) && ( !header_continued || i>0 ) ) {
	l=i+9; while( l<j && buf[l]!='\r' && buf[l]!='\n' ) l++;
	if( !printheaders ) { fwrite(buf+i,l-i,1,stderr); fprintf(stderr,"\n\n"); }
      }
      i=j+1;   // buf[j]=='\n', i<=len
    } while(i<len);
    if(break2) break;
    if( i > 0 ) {
      if(printheaders) fwrite(buf,i,1,printheaders==1?stdout:stderr);
      memmove(buf,buf+i,len-i);
      len=len-i; bufptr=buf+len; header_continued=false; 
    } else if( j >= BUFSZ ) {
      if(printheaders) fwrite(buf,BUFSZ,1,printheaders==1?stdout:stderr);
      bufptr=buf; len=0; header_continued=true;
    }
  }
  if( printheaders && j>0 ) fwrite(buf,j,1,printheaders==1?stdout:stderr);
  if( statuscode == 416 ) {
    if( is_continued && dld_pos > 0 ) { 
      statuscode = 200; fprintf(stderr,"apparently already fully downloaded: %s\n",urlspec->fullURL); 
      saveOutput = false; saveOutput_set = true;
    } else fprintf(stderr,"416 Requested Range Not Satisfiable\n");
  }
  //fprintf(stdout,"<%llu,%i>\n",content_length,had_content_length); printf("[%i,%i]\n",j,len);
  if(!saveOutput_set) {
    saveOutput = statuscode == 200 || statuscode == 206 || ( saveErrorResponse && !is_continued );
    if( saveOutput && !is_continued ) { 
      if(!OpenOutput(urlspec)) return FileErr;
  } }
  if( j < len ) {
    if(saveOutput) fwrite(buf+j,len-j,1,output);
    dld_pos += len-j; 
  }
  if(!quiet) InitProgress(dld_pos,content_length);
  while( !had_content_length || dld_pos < content_length ) {
    ioc = readin(fd,buf,min_macro(BUFSZ,content_length));
    if(ioc<0) { fprintf(stderr,"read error"); if(dldFileName) fclose(output); return ConnErr; }
    if(!ioc) { 
      if( had_content_length && dld_pos < content_length ) { fprintf(stderr,"short read: connection closed before end of file has been reached"); if(dldFileName) fclose(output); return false; }
      break;
    }
    if(saveOutput) fwrite(buf,ioc,1,output);
    dld_pos += ioc;
    if(!quiet) UpdateProgress(dld_pos,content_length);
  }
  if(!quiet) fputc('\n',stderr);
  if(saveOutput||saveOutput_set) RenameOutput();
  else if(is_continued) free(dldFileName);

  return errcode;
}

enum Command { CM_TII = 1, CM_TUU = 2, CM_TII_CERT = 3 } command;
enum Mode { MD_DEFAULT = 0, MD_DANE, MD_UNSECURE_DANE, MD_CERT, MD_NO_CHECK_CERT } mode = MD_DEFAULT;
bool force_ipv6 = false, save_rogue_cert = false; const char *certfile;
struct UrlSpec urlspec;

int main(int argc, char *argv[]) {
    int sd, i, errnum; int result = 0, curres; bool next_is_always_filename = false, opterr = false; bool resolver_inited=false;
    SSL_CTX *ctx; SSL *ssl; bool ctx_inited = false; X509 *conn_cert; void *tlsa_rec;

    if( argc <= 1 ) usage(true);
    if(!strcmp(argv[1],"--help")) usage(false);
    i=1;
    while( argv[i][0] == '-' ) { 
      if(!strcmp(argv[i],"--")) { i++; break; }
      if(!strcmp(argv[i],"--sys-keyfile")) { keyfile_mode |= KF_SYS; last_keyfile_mode = KF_SYS; }
      else if(!strcmp(argv[i],"--ub-keyfile")) { keyfile_mode |= KF_UNBOUND; last_keyfile_mode = KF_UNBOUND; }
      else if(!strcmp(argv[i],"--keyfile")) { 
	if(argc<=i+1) { fprintf(stderr,"--keyfile: missing argument\n"); return 1; }
	switch(last_keyfile_mode) {
	  case KF_SYS: sys_keyfile = argv[i+1]; break;
	  case KF_UNBOUND:unbound_keyfile = argv[i+1]; break;
	  default: fprintf(stderr,"unknown keyfile mode; exiting"); return 1;
	}
	i++;
      } else { fprintf(stderr,"unknown global option: %s\n",argv[i]); opterr=true; }
      i++;
    }
    if(!strcmp(argv[i],"tii")) command = CM_TII;
    else if(!strcmp(argv[i],"tuu")) command = CM_TUU;
    else if(!strcmp(argv[i],"tii-cert")) command = CM_TII_CERT;
    else { fprintf(stderr,"unknown command: %s\n\n",argv[i]); exit(1); }
    if( command == CM_TUU ) { fprintf(stderr,"not implemented: tuu(put)\n\n"); return 1; }
    i++;

    if(keyfile_mode) { InitResolver(); resolver_inited = true; }

    if( command == CM_TII_CERT ) {
      mode = MD_DANE; result = 0;

      while( argc > i ) {
	if( argv[i][0] == '-' && !next_is_always_filename ) {
	  if(argv[i][1] != '-' ) {
	    const char *oc;
	    for( oc=argv[i]+1; *oc; oc++ )
	      switch(*oc) {
		case 'u': save_rogue_cert = true; break;  case 'U': save_rogue_cert = false; break;
		case 'r': save_rogue_cert = true; break;  case 'R': save_rogue_cert = false; break;
		case 'q': quiet = true; if(verbosity>0) verbosity--; break;
		case 'v': verbosity++; break;
		default: fprintf(stderr,"unknown option for tii-cert: -%c\n",*oc); opterr=true; break;
	      }
	  } else {
	    if(!strcmp(argv[i],"--")) next_is_always_filename = true;
	    else { fprintf(stderr,"unknown option for tii-cert: %s\n",argv[i]); opterr=true; }
	  }
	} else {
	  if(opterr) return 1;  // { putchar('!'); 
	  if(parse_urlspec( argv[i], &urlspec, force_ipv6 )) return 1;
	  if(verbosity>=3) print_urlspec(stdout,&urlspec);
	  if( urlspec.protocol == PR_HTTP || urlspec.protocol == PR_FTP ) {
	    fprintf(stderr,"only https and ftps use certificates and encryption: %s",urlspec.fullURL);
	    continue;
	  }
	  sd = socket(urlspec.addr->ai_addr->sa_family, SOCK_STREAM, 0);
	  errnum = socket_connect(&sd,urlspec.addr,0); 
	  if(!errnum) {
	    if(!ctx_inited) { 
	      SSL_library_init();
	      ctx_inited = InitCTX(&ctx);
	    }
	    if(ctx_inited) {
	      ssl = SSL_new(ctx);      /* create new SSL connection state */
	      SSL_set_fd(ssl,sd);    /* attach the socket descriptor */
	      if ( SSL_connect(ssl) == -1 ) { ERR_print_errors_fp(stderr); result = ConnErr; }
	      else { 
		if(!resolver_inited) { InitResolver(); resolver_inited = true; }
		curres = retrieve_TLSA( urlspec.host, &tlsa_rec, &urlspec );
		conn_cert = SSL_get_peer_certificate(ssl);   /* get the server's certificate */
		if( tlsa_rec ) {
		  if( !curres || !verify_hash( tlsa_rec, conn_cert, &urlspec ) ) { 
		    if(save_rogue_cert) saveCert(urlspec.host,curres?"-rogue":"-no-DNSSEC",conn_cert);
		    result=2; 
		  } else saveCert(urlspec.host,"",conn_cert);
		} else { 
		  if(save_rogue_cert) saveCert(urlspec.host,"-not-validated",conn_cert); 
		  result=2; 
		}
	        SSL_free(ssl); 
	      }
	} } }
	i++;
      }

    } else {

      while( argc > i ) {
	if( argv[i][0] == '-' && !next_is_always_filename ) {
	  if(argv[i][1] != '-' ) {
	    const char *oc;
	    for( oc=argv[i]+1; *oc; oc++ )
	      switch(*oc) {
		case 'c': do_continue = true; break;
		case 'q': quiet = true; if(verbosity>0) verbosity--; break;
		case 'v': verbosity++; break;
		case 'd': mode = MD_DANE; break;
		case 'u': mode = MD_UNSECURE_DANE; break;
		case 'n': mode = MD_NO_CHECK_CERT; break;
		case 's': fileNameMode = FS_STDOUT; break;
		case '0': mode = MD_DEFAULT; break;
		case '6': force_ipv6 = true; break;
		case 'h': printheaders = 1; break;             case 'H': printheaders = 0; break;
		case 'o': overwriteTargetFile = true; break;   case 'O': overwriteTargetFile = false; break;
		case 'e': saveErrorResponse = true; break;     case 'E': saveErrorResponse = false; break;
		case 'r': save_rogue_cert = true; break;       case 'R': save_rogue_cert = false; break;
		default: fprintf(stderr,"unknown URL option: -%c\n",*oc); opterr=true; break;
	      }
	  } else {
	    if(!strcmp(argv[i],"--")) next_is_always_filename = true;
	    else if(!strcmp(argv[i],"--cert")) {
	      if(argc<=i+1) { fprintf(stderr,"--cert: missing argument\n"); return 1; }
	      mode = MD_CERT; certfile = argv[i+1]; i++;
	      if(required_cert) X509_free(required_cert);
	      required_cert = read_cert_from_file(certfile);
	    } else if(!strcmp(argv[i],"--dane")) mode = MD_DANE;
	    else if(!strcmp(argv[i],"--unsecure-dane")) mode = MD_UNSECURE_DANE;
	    else if(!strcmp(argv[i],"--no-check-cert")) mode = MD_NO_CHECK_CERT;
	    else if(!strcmp(argv[i],"--to-stdout")) fileNameMode = FS_STDOUT;
	    else if(!strcmp(argv[i],"--help")) usage(false);
	    else if(!strcmp(argv[i],"--to")) {
	      if(argc<=i+1) { fprintf(stderr,"--to: missing argument\n"); return 1; }
	      fileNameMode = FS_EXPLICIT;
	      fileName = argv[i+1]; i++;
	    } else { fprintf(stderr,"unknown URL option: %s\n",argv[i]); opterr=true; }
	  }
	} else {
	  if(opterr) return 1;  // { putchar('!'); 
	  if(parse_urlspec( argv[i], &urlspec, force_ipv6 )) return 1;
	  if(verbosity>=3) print_urlspec(stdout,&urlspec);
	  sd = socket(urlspec.addr->ai_addr->sa_family, SOCK_STREAM, 0);
	  errnum = socket_connect(&sd,urlspec.addr,0); 
	  if(!errnum) {
	    switch(urlspec.protocol) {
	      case PR_HTTP:
		if( mode == MD_CERT || mode == MD_DANE || mode == MD_UNSECURE_DANE ) { fprintf(stderr,"http does not support encryption, certificates and DANE - not downloading."); break; }
		curres = mk_HTTP_Request(&urlspec,&sd,directRead,directWrite,false);
		if( curres && curres < result ) result = curres;
		break;
	      case PR_HTTPS:
		if( mode == MD_CERT && !required_cert ) abort();
		if(!ctx_inited) { 
		  SSL_library_init();
		  ctx_inited = InitCTX(&ctx);
		}
		if(ctx_inited) {
		  ssl = SSL_new(ctx);      /* create new SSL connection state */
		  SSL_set_fd(ssl,sd);    /* attach the socket descriptor */
		  if ( SSL_connect(ssl) == -1 ) { ERR_print_errors_fp(stderr); if(result < ConnErr) result = ConnErr; }
		  else { 
		    if( mode == MD_CERT ) {
		      conn_cert = SSL_get_peer_certificate(ssl);   /* get the server's certificate */
		      if(!conn_cert) { fprintf(stderr,"failed to retrieve server certificate; closing connection (%s).\n",argv[i]); break; }
		      if(!X509_compare_equal(conn_cert,required_cert)) { 
			fprintf(stderr,"server certificate does not match the required certificate; closing connection (%s).\n",argv[2]); 
			if(save_rogue_cert) saveCert(urlspec.host,"-rogue",conn_cert);
			break; 
		      }
		      X509_free(conn_cert);
		    } else if( mode == MD_DANE || mode == MD_UNSECURE_DANE || mode == MD_DEFAULT ) {
		      if(!resolver_inited) { InitResolver(); resolver_inited = true; }
		      if( !retrieve_TLSA( urlspec.host, &tlsa_rec, &urlspec ) && mode!=MD_UNSECURE_DANE ) break;
		      if(!tlsa_rec) break;
		      conn_cert = SSL_get_peer_certificate(ssl);   /* get the server's certificate */
		      if(!verify_hash( tlsa_rec, conn_cert, &urlspec )) {
			if(save_rogue_cert) saveCert(urlspec.host,"-rogue",conn_cert);
			break;
		    } }
		    curres = mk_HTTP_Request(&urlspec,ssl,(PReadIn)SSL_read,(PWriteOut)SSL_write,true); 
		    if( curres && curres > result ) result = curres;
		    SSL_free(ssl); 
		} }
		break;
	    }
	    close(sd);         /* close socket */
	  }
	  fileNameMode = FS_IMPLICIT;
	  next_is_always_filename = false;
	}
	i++;
    } }
    if(ctx_inited) SSL_CTX_free(ctx);
    if(required_cert) X509_free(required_cert);
    if(resolver_inited) CleanUpResolver();
    return result;
}
