#ifndef _DANE_SUPPORT_H
#define _DANE_SUPPORT_H

#include <openssl/ssl.h>
#include "basedefines.h"

// resolv.conf:
//  Cloudflare: 1.1.1.1, 1.0.0.1
//  Google: 8.8.8.8, 8.8.4.4
//  Quad: 9.9.9.9

#define SYS_KEYFILE "/usr/share/dns/root.key"
#define SYS_KEYFILE_2 "/etc/trusted-key.key"
#define UB_KEYFILE "/etc/unbound/root.key"

#define DF_NSFILE_1 "/etc/resolv.conf.DANE"
#define DF_NSFILE_2 "/etc/resolv.conf"

#ifndef T_TLSA
  #define T_TLSA 52
#endif

typedef unsigned char uchar;

enum NSFileMode { NS_DEFAULT, NS_EXPLICIT };
extern enum NSFileMode nsfile_mode;
extern const char *nsfile;   // for NS_EXPLICIT only, default is NULL

enum KeyFileMode { KF_DEFAULT = 0, KF_UNBOUND = 1, KF_SYS = 2 };   // default is effectively the same as unbound
extern enum KeyFileMode keyfile_mode;  // in deed a bit array though normally only one bit will be set
extern enum KeyFileMode last_keyfile_mode;
extern const char *unbound_keyfile, *sys_keyfile;  // set to USB_ & SYS_KEYFILE by default 

enum ShaType { ST_NONE = 0, ST_SHA256 = 1, ST_SHA512 = 2, ST_count = 3 };  // ShaType, equals number in TLSA record
enum ShaWhatOfCert { SK_FULLCERT = 0, SK_PUBKEY = 1 };   // ShaKind, equals number in TLSA record
extern int shalen[ST_count];
bool sha_of_cert( uchar *digest, int shaType, int shaWhat, X509 *cert, struct UrlSpec *urlspec );  // prints error messages

bool InitResolver(); void CleanUpResolver();
bool retrieve_TLSA( const char *domain, void **tlsa_rec, struct UrlSpec *urlspec );
bool verify_hash( void *TLSArec, X509 *cert, struct UrlSpec *urlspec );  // prints error messages

#endif
