
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <arpa/inet.h> 
//#include <resolv.h>
#include <netdb.h>
#include <unbound.h>

#include <sys/types.h> 
#include <sys/stat.h> 

#include <unistd.h>
#include <alloca.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include "sockcommon.h"
#include "dane-support.h"
// --

//#define SYS_KEYFILE "/usr/share/dns/root.key"
//#define SYS_KEYFILE_2 "/etc/trusted-key.key"
//#define UB_KEYFILE "/etc/unbound/root.key"

//#define DF_NSFILE_1 "/etc/resolv.conf.DANE"
//#define DF_NSFILE_2 "/etc/resolv.conf"

//enum NSFileMode { NS_DEFAULT, NS_EXPLICIT };
enum NSFileMode nsfile_mode = NS_DEFAULT;
const char *nsfile = NULL;   // for NS_EXPLICIT only

// enum KeyFileMode { KF_DEFAULT = 0, KF_UNBOUND = 1, KF_SYS = 2 };   // default is effectively the same as unbound
enum KeyFileMode keyfile_mode = KF_DEFAULT;  // set to zero, bit array
enum KeyFileMode last_keyfile_mode = KF_UNBOUND;  // used to see of which type a keyfile stated as parameter will be
const char *unbound_keyfile = UB_KEYFILE, *sys_keyfile = SYS_KEYFILE; 

//enum ShaType { ST_SHA256, ST_SHA512 };
//enum ShaWhatOfCert { SK_FULLCERT, SK_PUBKEY };
int shalen[ST_count] = { -1, 32, 64 };

const EVP_MD *md_sha256;
const EVP_MD *md_sha512;

bool sha_of_cert( uchar *digest, int shaWhat, int shaType, X509 *cert, struct UrlSpec *urlspec ) {
  X509_PUBKEY *pubkey; uchar *der_pubkey = NULL; int der_len; unsigned int digest_len; const EVP_MD *md;
  switch(shaWhat) {
    case SK_PUBKEY:
      pubkey = X509_get_X509_PUBKEY(cert); if(!pubkey) return false;
      der_len = i2d_X509_PUBKEY(pubkey,&der_pubkey);
      X509_PUBKEY_free(pubkey);
      if(der_len>=0) switch(shaType) {
	case ST_SHA256: SHA256( der_pubkey, der_len, digest ); break;
	case ST_SHA512: SHA512( der_pubkey, der_len, digest ); break;
	default: fprintf(stderr,"DANE: sha_of_cert: ShaType %i not implemented (%s)\n",shaType,urlspec->fullURL); return false;
      }
      OPENSSL_free(der_pubkey);
      return der_len >= 0;
    case SK_FULLCERT:
      switch(shaType) { case ST_SHA256: md = md_sha256; break; case ST_SHA512: md = md_sha512; break; default: fprintf(stderr,"DANE: sha_of_cert: ShaType %i not implemented (%s)\n",shaType,urlspec->fullURL); return false; }
      return X509_digest(cert, md, digest, &digest_len);
    default: fprintf(stderr,"DANE: sha_of_cert: unknown ShaWhat mode (%i) - %s\n",shaWhat, urlspec->fullURL); return false;
  }
  return true;
}

struct ub_ctx * ctx = NULL;

bool InitResolver() {
  const char *eff_nsfile; int retval; struct stat statbuf;
  md_sha256 = EVP_sha256(); md_sha512 = EVP_sha512();
  ctx = ub_ctx_create(); if(!ctx) { fputs("DANE: failed to create unbound-context\n",stderr); return false; }
  switch(nsfile_mode) {
    case NS_DEFAULT:
      if(stat(DF_NSFILE_1,&statbuf)==0) eff_nsfile = DF_NSFILE_1; else eff_nsfile = DF_NSFILE_2;
      break;
    case NS_EXPLICIT: eff_nsfile = nsfile; break;
    default: fprintf(stderr,"DANE: InitResolver: unknown nsfile_mode\n"); abort();
  }
  if( ( retval = ub_ctx_resolvconf(ctx,eff_nsfile) ) != 0 ) { printf("DANE: error reading resolv.conf: %s; errno says: %s\n", ub_strerror(retval), strerror(errno)); return false; }
  if( keyfile_mode & KF_SYS )
    if( ( retval = ub_ctx_add_ta_file(ctx, sys_keyfile) ) != 0 ) { fprintf(stderr,"DANE: error adding public DNSSEC keys: %s; errno says: %s\n", ub_strerror(retval),strerror(errno)); return false; }
  if(( keyfile_mode & KF_UNBOUND ) || keyfile_mode == KF_DEFAULT )
    if( ( retval = ub_ctx_trustedkeys(ctx, unbound_keyfile) ) != 0 ) { fprintf(stderr,"DANE: error adding public DNSSEC keys: %s; errno says: %s\n", ub_strerror(retval),strerror(errno)); return false; }
  return true;
}

bool retrieve_TLSA( const char *domain, void **tlsa_rec, struct UrlSpec *urlspec ) {
  int len = strlen(domain); int bufsz; char *domainIdent = alloca(bufsz=len+12);
  int retval; struct ub_result *result; 
  snprintf(domainIdent,bufsz,"_443._tcp.%s",domain);
  retval = ub_resolve( ctx, domainIdent, T_TLSA, C_IN, &result );
  if( retval==0 && result->havedata ) 
    *tlsa_rec = result;
  else { 
    if(retval!=0) fprintf(stderr,"DANE: resolver error: %s (%s).\n",ub_strerror(retval),urlspec->fullURL); else fprintf(stderr,"DANE: no data (%s)\n",urlspec->fullURL); 
    *tlsa_rec = NULL; return false; 
  }
  if(!result->secure) {
    fprintf(stderr,"DANE: TLSA record of %s not validated as secure: ",domain);
    if(result->bogus) fprintf(stderr,"a security failure! (bogus because %s) - %s.\n", result->why_bogus, urlspec->fullURL );
		 else fprintf(stderr,"not a security failure (not bogus) - %s.\n",urlspec->fullURL);
    return false;
  }
  return true;
}

bool verify_hash( void *TLSArec_asvoid, X509 *cert, struct UrlSpec *urlspec ) {
  int *l, i; char **thisdata; bool result = false, retval; 
  int lenA; uchar *bufA; X509_PUBKEY *pubkey;
  struct ub_result *TLSArec = (struct ub_result*) TLSArec_asvoid; 
  uchar digest[64];  // enough for SHA512
  for( l=TLSArec->len, thisdata=TLSArec->data; *l && *thisdata; l++, thisdata++ ) {
    // result len=35; 03:00:01:a8:ed:f0:ca:ca:f7:76:ac:ac:df:e5:35:64:c5:15:56:ad:32:5f:03:a3:69:e4:c8:f4:62:2b:4d:c5:b0:68:65:,
    //for( i = 0; i < *l; i++ ) printf("%02x:",(int)(unsigned char)(*thisdata)[i]);  // ignore three bytes
    if( (*thisdata)[2] != ST_NONE ) {
      retval = sha_of_cert( digest, (*thisdata)[1], (*thisdata)[2], cert, urlspec );
      if( *l != shalen[(int)(*thisdata)[2]] + 3 ) fprintf(stderr,"DANE: warning: Length of TLSA record not as expected! It may contain invalid information (%s).\n",urlspec->fullURL);
      if(!retval||shalen[(int)(*thisdata)[2]]>*l-3) continue;
      if(!memcmp(*thisdata+3,digest,shalen[(int)(*thisdata)[2]])) { result = true; break; }
      if( !result && verbosity>=2 ) {
	fprintf(stderr,"TLSA record (first three bytes are for TLSA-mode): ");
	for( i = 0; i < *l; i++ ) fprintf(stderr,"%s%02x",i?":":"",(int)(unsigned char)(*thisdata)[i]); fputc('\n',stderr);   // ignore three bytes
	fprintf(stderr,"hash of the server certificate: ");
	for( i = 0; i < shalen[(int)(*thisdata)[2]]; i++ ) fprintf(stderr,"%s%02x",i?":":"",(int)(unsigned char)digest[i]); fputc('\n',stderr);  // ignore three bytes
      }
    } else {
      switch((*thisdata)[1]) {
        case SK_FULLCERT:
	  lenA = i2d_X509(cert,&bufA);
	  if( lenA == *l - 3  && !memcmp(*thisdata+3,bufA,lenA) ) result = true;
	  OPENSSL_free(bufA);
	  break;
	case SK_PUBKEY:
	  pubkey = X509_get_X509_PUBKEY(cert);
	  lenA = i2d_X509_PUBKEY(pubkey,&bufA);
	  X509_PUBKEY_free(pubkey);
	  if( lenA == *l - 3  && !memcmp(*thisdata+3,bufA,lenA) ) result = true;
	  OPENSSL_free(bufA);
	  break;
	default:
	  fprintf(stderr,"DANE: verify_hash: unknown ShaWhat mode (%s)\n",urlspec->fullURL);
	  break;
      }
      if(result) break;
    } 
  }
  ub_resolve_free(TLSArec);
  if(!result) fprintf(stderr,"certificate signature in TLSA record did not match (%s)\n",urlspec->fullURL);
  return result;
}

void CleanUpResolver() {
  if(ctx) ub_ctx_delete(ctx);
  ctx = NULL;
}

