//
//  sctp helper functions
//   this code may be used under the terms of GPLv3
//   if you apply changes please sign the contributor license aggreement at https://www.elstel.org/mohican/ 
//   so that your changes can be included into the main trunk, (c) copyright by Elmar Stellnberger
//

#ifndef _SOCK_COMMON_H
#define _SOCK_COMMON_H
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>	  // in add. to pthread.h: EOWNERDEAD return value
#include <stdio.h>
#include "basedefines.h"

#define aux_sysmsg(message) fputs(message,stderr);
#define aux_error(message) aux_sysmsg(message)
#define aux_warning(message) aux_sysmsg(message)
#define aux_sysmsg_int(value) fprintf(stderr,"%lli",(long long)value);
#define aux_sysmsg_adr(value) fprintf(stderr,"%llx",(unsigned long long)value);

#define aux_praefixed_assert(condition,praefix,message) if(unlikely(!(condition))) { aux_error(praefix "assert failed at " __FILE__ ":"); aux_sysmsg_int(__LINE__); aux_sysmsg(": " message #condition "\n\n"); abort(); }
#define aux_assert_(condition,message) aux_praefixed_assert(condition,"aux_",message " ~ ")
#define aux_assert(condition) aux_praefixed_assert(condition,"aux_","")

extern int verbosity;   // default is 1

int setmicroseconds( struct timeval *tv, const char *mstxt, long jiffie );
int setnanoseconds( struct timespec *ts, const char *mstxt, long jiffie );
// pass 0 or a negative value for jiffie to accept any timespan; otherwise jiffie gives the minimum frequency

int doclose(int fd);
void dup2quitonerr( int oldfd, int newfd );

bool checkascii( const char *str );

// priamry sending, receival and temporary buffer
#define SF_ANYHOST 1
#define SF_IPV6 2
#define SF_SILENT 4

enum Protocol { PR_HTTP = 1, PR_HTTPS = 2, PR_FTP = 3, PR_FTPS = 4 };
extern const char* protocol2str[3];

struct UrlSpec {
  struct addrinfo *addr; int protocol, port_override;  // flags: 1~anyhost, 2~ipv6
  const char *host, *path, *target, *fullURL;
};

int parse_urlspec( char *url, struct UrlSpec *urlspec, bool force_ipv6 );
void print_sockaddr( FILE* out, struct sockaddr *sa );
void print_urlspec( FILE* out, struct UrlSpec *urlspec );

// flags: pos0: anyhost; pos1: ipv6; always frees listening_host_ais; uses p2error & printerr (i.e. logerr must be initialized)
int socket_connect( int *socket, struct addrinfo *listening_host_ais, int flags ); 

#if OSTYPE & OS_linux
  #define OSSTR " (linux-gnu)"
#elif OSTYPE & OS_OpenBSD
  #define OSSTR " (OpenBSD)"
#elif OSTYPE & OS_BSD
  #define OSSTR " (BSD)"
#elif OSTYPE & OS_windows
  #define OSSTR " (Windows)"
#elif OSTYPE & OS_macos
  #define OSSTR " (MacOS)"
#elif OSTYPE & OS_os2
  #define OSSTR " (OS/2)"
#elif OSTYPE & OS_unix
  #define OSSTR " (Unix)"
#elif OSTYPE & OS_dos
  #define OSSTR " (DOS)"
#else
  #define OSSTR ""
#endif


#endif
